# -*- coding: utf-8 -*-

"""
    Unit tests for the divergence module of the SmartSVM package.
"""

from __future__ import division

import unittest
import numpy as np

from smartsvm import divergence


class MstTestCase(unittest.TestCase):

    def test_merge_and_label(self):
        """ DIVERGENCE: Test merging and labelling of datasets """
        X = np.array([
            [0, 1, 2],
            [3, 2, 1],
            ])
        Y = np.array([
            [9, 8, 7],
            [6, 5, 4],
            [0, 1, 0]
            ])
        data, labels = divergence.merge_and_label(X, Y)
        self.assertTrue(np.array_equal(data, np.array([
            [0, 1, 2],
            [3, 2, 1],
            [9, 8, 7],
            [6, 5, 4],
            [0, 1, 0]
            ])))
        self.assertTrue(np.array_equal(labels,
                        np.array([-1., -1., 1., 1., 1.])))

    def test_divergence_correction_1(self):
        """ DIVERGENCE: Testing divergence measure on small dataset. """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            ])
        Y = np.array([
            [4, 1],
            [3, -3],
            [5, 4]
            ])
        div = divergence.divergence(X, Y)
        self.assertAlmostEqual(div, 1.0/3.0)

    def test_divergence_no_correction_1(self):
        """ DIVERGENCE: Testing divergence measure on small dataset. """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            ])
        Y = np.array([
            [4, 1],
            [3, -3],
            [5, 4]
            ])
        div = divergence.divergence(X, Y, bias_correction=False)
        self.assertAlmostEqual(div, 1.0/3.0)




if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Useful numerical utilities.
"""

from __future__ import division

import numpy as np

def indices_from_classes(classes, y):
    """Create index vector for all labels in classes

    Create an index vector with the same dimensions as the vector y, with 
    indices equal to True if the label is in the list of classes.

    Parameters
    ----------
    classes : list or numpy array
        Classes to check membership of

    y : numpy.ndarray
        Label vector

    Returns
    -------
    indices : numpy.ndarray (bool)
        Boolean vector with True for the elements of ``y`` which occur in 
        ``classes`` and False otherwise.

    """
    indices = np.zeros(y.shape, dtype=bool)
    for c in classes:
        indices = np.logical_or(indices, y == c)
    return indices



"""
    Unit tests for the error estimate module of the SmartSVM package.
"""

from __future__ import division

import math
import unittest

import numpy as np
import networkx as nx

from smartsvm import error_estimate


class ErrorEstimateTestCase(unittest.TestCase):

    def test_worker_1(self):
        """ ERROR_ESTIMATE: Test worker """
        # This uses the same dataset in the test_divergence code, so we know 
        # that the divergence is 1/3, so we can calculate what the error 
        # estimate should be.
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            [4, 1],
            [3, -3],
            [5, 4]
            ])
        y = np.array([1, 1, 1, 2, 2, 2])
        i = 1
        j = 2
        normalize = False
        Xyijn = (X, y, i, j, normalize)
        exp = (i, j, 1/2 - 1/4 * math.sqrt(1/3) - 1/4*1/3)
        res = error_estimate._worker(Xyijn)
        self.assertEqual(exp[0], res[0])
        self.assertEqual(exp[1], res[1])
        self.assertAlmostEqual(exp[2], res[2])

    def test_compute_error_graph(self):
        """ ERROR ESTIMATE: Test build error graph """
        # Test using a very simple dataset for which all pairwise errors are 
        # the same, equal to 1/2 - 1/4*sqrt(1/3) - 1/4*1/3
        X = np.array([
            [-0.05, 3],
            [0.05, 3],
            [0, 3.00],
            [3, 0.05],
            [3, -0.05],
            [3.00, 0],
            [-0.05, -3],
            [0.05, -3],
            [0, -3.00],
            [-3, 0.05],
            [-3, -0.05],
            [-3.00, 0],
            ])
        y = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])
        # divide by 0.5 since we use the normalized error
        exp_weight = ((1.0/2.0 - 1.0/4.0 * math.sqrt(1.0/3.0) - 
            1.0/4.0*1.0/3.0) / 0.5)
        exp_G = nx.Graph()
        exp_G.add_edge(1, 2, weight=exp_weight)
        exp_G.add_edge(1, 3, weight=exp_weight)
        exp_G.add_edge(1, 4, weight=exp_weight)
        exp_G.add_edge(2, 3, weight=exp_weight)
        exp_G.add_edge(2, 4, weight=exp_weight)
        exp_G.add_edge(3, 4, weight=exp_weight)

        res_G_1 = error_estimate.compute_error_graph(X, y, n_jobs=1)
        em = nx.algorithms.isomorphism.numerical_edge_match('weight', 1)
        self.assertTrue(nx.is_isomorphic(res_G_1, exp_G, edge_match=em))
        self.assertEqual(sorted(res_G_1.nodes()), sorted(exp_G.nodes()))

        res_G_2 = error_estimate.compute_error_graph(X, y, n_jobs=2)
        em = nx.algorithms.isomorphism.numerical_edge_match('weight', 1)
        self.assertTrue(nx.is_isomorphic(res_G_2, exp_G, edge_match=em))
        self.assertEqual(sorted(res_G_2.nodes()), sorted(exp_G.nodes()))

    def test_compute_ovr_error_normalize(self):
        """ ERROR ESTIMATE: Test build error graph OvR (normalized)"""
        # Test using a very simple dataset. Error rates are checked using 
        # Matlab.
        X = np.array([
            [-0.05, 3],
            [ 0.05, 3],
            [ 0.00, 3],
            [3.1,  0.05],
            [3.1, -0.05],
            [3.1,  0.00],
            [-0.05, -3.3],
            [ 0.05, -3.3],
            [ 0.00, -3.3],
            [-3.7,  0.05],
            [-3.7, -0.05],
            [-3.7,  0.00],
            ])
        y = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])
        N = X.shape[0]

        v1 = (1.0/2.0 - 1.0/4.0*math.sqrt(0.381966011250105) - 1.0/4.0 * 
                0.381966011250105) / min(3/N, (N-3)/N)
        v2 = (1.0/2.0 - 1.0/4.0*math.sqrt(0.388888888888889) - 1.0/4.0 * 
                0.388888888888889) / min(3/N, (N-3)/N)
        v3 = (1.0/2.0 - 1.0/4.0*math.sqrt(0.833333333333333) - 1.0/4.0 * 
                0.833333333333333) / min(3/N, (N-3)/N)
        v4 = (1.0/2.0 - 1.0/4.0*math.sqrt(0.777777777777778) - 1.0/4.0 * 
                0.777777777777778) / min(3/N, (N-3)/N)

        exp = {1: v1, 2: v2, 3: v3, 4: v4}

        res_1 = error_estimate.compute_ovr_error(X, y)
        res_2 = error_estimate.compute_ovr_error(X, y)
        for i in range(1, 5):
            self.assertAlmostEqual(exp[i], res_1[i])
            self.assertAlmostEqual(exp[i], res_2[i])

    def test_compute_ovr_error_no_normalize(self):
        """ ERROR ESTIMATE: Test build error graph OvR (not normalized)"""
        # Test using a very simple dataset. Error rates are checked using 
        # Matlab.
        X = np.array([
            [-0.05, 3],
            [ 0.05, 3],
            [ 0.00, 3],
            [3.1,  0.05],
            [3.1, -0.05],
            [3.1,  0.00],
            [-0.05, -3.3],
            [ 0.05, -3.3],
            [ 0.00, -3.3],
            [-3.7,  0.05],
            [-3.7, -0.05],
            [-3.7,  0.00],
            ])
        y = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])

        exp = {
                1: (1.0/2.0 - 1.0/4.0*math.sqrt(0.381966011250105) - 1.0/4.0 * 
                    0.381966011250105),
                2: (1.0/2.0 - 1.0/4.0*math.sqrt(0.388888888888889) - 1.0/4.0 * 
                    0.388888888888889),
                3: (1.0/2.0 - 1.0/4.0*math.sqrt(0.833333333333333) - 1.0/4.0 * 
                    0.833333333333333),
                4: (1.0/2.0 - 1.0/4.0*math.sqrt(0.777777777777778) - 1.0/4.0 * 
                    0.777777777777778)
        }

        res_1 = error_estimate.compute_ovr_error(X, y, normalize=False)
        res_2 = error_estimate.compute_ovr_error(X, y, normalize=False)
        for i in range(1, 5):
            self.assertAlmostEqual(exp[i], res_1[i])
            self.assertAlmostEqual(exp[i], res_2[i])



    def test_hp_estimate_1(self):
        """ ERROR ESTIMATE: Test error estimate even unnormalized """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            ])
        Y = np.array([
            [4, 1],
            [3, -3],
            [5, 4]
            ])

        exp = 5.0/12.0 - 1.0/(4.0*math.sqrt(3.0))
        res = error_estimate.hp_estimate(X, Y, normalize=False)
        self.assertEqual(res, exp)

    def test_hp_estimate_2(self):
        """ ERROR ESTIMATE: Test error estimate uneven unnormalized """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            ])
        Y = np.array([
            [4, 1],
            [3, -3],
            [5, 4],
            [3, 3]
            ])

        exp = 0.22919372518014852
        res = error_estimate.hp_estimate(X, Y, normalize=False)
        self.assertEqual(res, exp)

    def test_hp_estimate_3(self):
        """ ERROR ESTIMATE: Test error estimate even normalized """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            ])
        Y = np.array([
            [4, 1],
            [3, -3],
            [5, 4],
            ])

        exp = 5.0/12.0 - 1.0/(4.0*math.sqrt(3.0))
        exp = exp * 2.0
        res = error_estimate.hp_estimate(X, Y, normalize=True)
        self.assertEqual(res, exp)

    def test_hp_estimate_4(self):
        """ ERROR ESTIMATE: Test error estimate uneven normalized """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            ])
        Y = np.array([
            [4, 1],
            [3, -3],
            [5, 4],
            [3, 3]
            ])

        exp = 0.22919372518014852
        exp = exp / (3.0 / 7.0)
        res = error_estimate.hp_estimate(X, Y, normalize=True)
        self.assertEqual(res, exp)


if __name__ == '__main__':
    unittest.main()



"""
    Unit tests for the utilities of the SmartSVM package
"""

from __future__ import division

import numpy as np
import unittest

from smartsvm import utils

class UtilsTestCase(unittest.TestCase):

    def test_indices_from_classes(self):
        """ UTILS: Test creating index vector from y and classes set """
        classes = set([1, 2, 3])
        y = np.array([1, 1, 2, 4, 5, 3, 2, 3, 5, 4, 1, 6])
        exp = np.array([True, True, True, False, False, True ,True ,True, 
            False, False, True, False])
        res = utils.indices_from_classes(classes, y)
        for e, r in zip(exp, res):
            self.assertEqual(e, r)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Functions for graph cutting

This module contains functions for cutting the weighted error graph. The 
default algorithm for cutting the graph is the Stoer-Wagner algorithm, but this 
module also contains code for experimenting with the Spectral Clustering 
algorithm to create graph cuts.

"""

from __future__ import division

import networkx as nx
import numpy as np

from collections import Counter

from sklearn.cluster import spectral_clustering
from scipy.sparse.csgraph import connected_components


def graph_cut(G, algorithm='stoer_wagner'):
    """Cut a connected graph into two disjoint graphs

    This is a wrapper function to make it easy to use different graph cut 
    algorithms.

    Parameters
    ----------
    G : networkx.Graph
        The graph that needs to be cut

    algorithm : str
        The graph cut algorithm to use. Available values are: 
        ``'stoer_wagner'``, ``'spectral_clustering'``, and 
        ``'normalized_cut'``.

    Returns
    -------
    G1 : networkx.Graph
        One of the subgraphs

    G2 : networkx.Graph
        The other subgraph

    Raises
    ------
    ValueError
        When an unknown graph cut algorithm is supplied.

    """
    if algorithm == 'stoer_wagner':
        return graph_cut_sw(G)
    elif algorithm == 'spectral_clustering':
        return graph_cut_sc(G, assign_labels='kmeans')
    elif algorithm == 'normalized_cut':
        return graph_cut_sc(G, assign_labels='discretize')
    else:
        raise ValueError("Unknown graph cut algorithm supplied: %s" %
                         algorithm)


def graph_cut_sw(G):
    """Apply the Stoer-Wagner graph cut

    Use the Stoer-Wagner algorithm for cutting the graph.

    Parameters
    ----------
    G : networkx.Graph
        The graph that needs to be cut

    Returns
    -------
    G1 : networkx.Graph
        One of the subgraphs

    G2 : networkx.Graph
        The other subgraph


    """
    cut_value, partition = nx.stoer_wagner(G)
    G1 = nx.subgraph(G, partition[0])
    G2 = nx.subgraph(G, partition[1])
    return G1, G2


def graph_cut_sc(G, assign_labels=None):
    """Apply the Spectral Clustering algorithm to cut the graph

    Create two subgraphs using a Spectral Clustering of the adjacency matrix of 
    the graph.

    Important: Since the matrix of weights is a dissimilarity matrix (high 
    numbers correspond to difficult to separate classes, we turn it into a 
    similarity matrix for the Spectral Clustering algorithm by using the 
    normalized exponent of the weight matrix. This is also done in the examples 
    of Scikit-Learn for Spectral Clustering. Weights that were zero in the 
    weight matrix are set to zero in the dissimilarity matrix as well.

    Parameters
    ----------
    G : networkx.Graph
        The graph that needs to be cut

    assign_labels : str
        Parameter for the Scikit-Learn ``spectral_clustering`` function.  
        Available values are: ``kmeans`` and ``discretize``.


    Returns
    -------
    G1 : networkx.Graph
        One of the subgraphs

    G2 : networkx.Graph
        The other subgraph


    """

    nodes = np.array(G.nodes())
    weights = np.array(nx.to_numpy_matrix(G))
    if weights.std() == 0.:
        data = weights.copy()
    else:
        data = np.exp(-weights/weights.std())
        data[weights == 0] = 0

    n_components, component_idx = connected_components(data)
    if n_components == 1:
        if len(nodes) > 2:
            labels = spectral_clustering(data, n_clusters=2, 
                    assign_labels=assign_labels)
        else:
            labels = np.array([0, 1])

        partition_0 = nodes[labels == 0]
        partition_1 = nodes[labels == 1]
    elif n_components == 2:
        partition_0 = nodes[component_idx == 0]
        partition_1 = nodes[component_idx == 1]
    else:
        most_common = Counter(component_idx).most_common(1)[0][0]
        partition_0 = nodes[component_idx == most_common]
        partition_1 = nodes[component_idx != most_common]

    G1 = nx.subgraph(G, partition_0)
    G2 = nx.subgraph(G, partition_1)

    return G1, G2



# -*- coding: utf-8 -*-

"""Functions for computing the Henze-Penrose estimate of the Bayes Error Rate

This module contains functions for computing the Henze-Penrose estimate of the 
Bayes Error Rate (BER).

"""

from __future__ import division

import numpy as np

from itertools import combinations
from math import sqrt
from multiprocessing import Pool
from networkx import Graph

from .divergence import divergence, ovr_divergence


def _worker(Xyijn):
    """Worker code for pairwise error estimation

    Calculates the error estimate for a pair of classes. The input Xyij is a 
    tuple with (X, y, i, j). The error is estimated between class i and class 
    j, where the data for classes i and j is gathered using ``X[y == i, :]``.

    Parameters
    ----------
    Xyijn : tuple
        Tuple with ``(X, y, i, j, normalize)`` where ``X`` is the dataset 
        containing at least the data for classes ``i`` and ``j``, ``y`` is the 
        vector of class labels, and ``normalize`` is a boolean variable 
        indicating if the normalized Bayes error estimates need to be 
        calculated.

    Returns
    -------
    tuple
        Tuple with (i, j, error)
    """
    X, y, i, j, norm = Xyijn
    Xi = X[y == i, :]
    Xj = X[y == j, :]
    error = hp_estimate(Xi, Xj, normalize=norm)
    return (i, j, error)


def compute_error_graph(X, y, n_jobs=1, normalize=True):
    """Compute the complete weighted graph of pairwise BER estimates

    Computes the estimate of the BER for each pair of classes and returns this 
    in a complete weighted graph. If desired, computing the error can be done 
    in parallel.

    Parameters
    ----------
    X : numpy.ndarray
        Numpy array of the data matrix

    y : numpy.ndarray
        Numpy array with the class labels

    n_jobs : int
        Number of parallel jobs to use

    normalize : bool
        Whether or not to use normalized BER estimation

    Returns
    -------
    G : networkx.Graph
        Weighted complete graph where the class labels are the nodes and the 
        edge weights are given by the BER estimate

    """
    classes = np.unique(y)

    # When n_jobs = 1 we don't want to use a pool of workers. This is useful 
    # when training using GridSearchCV from sklearn, since there it is not 
    # allowed for a child of GridSearchCV to have children of its own.
    args = [(X, y, i, j, normalize) for i, j in combinations(classes, 2)]
    if n_jobs == 1:
        results = [_worker(arg) for arg in args]
    else:
        pool = Pool(n_jobs)
        results = pool.map(_worker, args)
        pool.close()

    G = Graph()
    for i, j, weight in results:
        G.add_edge(i, j, weight=weight)

    return G


def compute_ovr_error(X, y, normalize=True):
    """Compute OvR-BER for each class

    The One-vs-Rest Bayes error rate is the error rate for a single class 
    compared to all other classes combined. This function computes this error 
    rate for each class in the dataset, using the 
    :func:`.divergence.ovr_divergence` function. By default the OvR-BER is 
    normalized with the estimates of the class prior probabilities.

    Parameters
    ----------
    X : numpy.ndarray
        The data matrix

    y : numpy.ndarray
        A vector of class labels

    normalize : bool
        Whether or not to normalize the OvR-BER

    Returns
    -------
    estimates : dict
        Dictionary with a mapping from the class label to a float representing 
        the OvR-BER estimate.

    """
    divergences = ovr_divergence(X, y)
    estimates = {k : 1.0/2.0 - 1.0/4.0*sqrt(dp) - 1.0/4.0*dp for k, dp in 
            divergences.items()}
    if normalize:
        for k in estimates:
            phat = X[y == k, :].shape[0] / (X.shape[0])
            qhat = X[y != k, :].shape[0] / (X.shape[0])
            ratio = min(phat, qhat)
            estimates[k] = estimates[k] / ratio
    return estimates


def hp_estimate(X1, X2, normalize=True):
    """Henze-Penrose estimation of the Bayes Error Rate

    Estimate the (normalized) Bayes error rate using the Henze-Penrose 
    estimator. The estimate is formed by averaging the upper and lower bounds 
    on the Bayes error. By default, the estimate is normalized with the 
    empirical estimates of the class prior probability.

    Parameters
    ----------
    X1 : numpy.ndarray
        Data matrix for one class

    X2 : numpy.ndarray
        Data matrix for the other class

    normalize : bool
        Whether or not to normalize the error using empirical estimates of 
        prior probabilities

    Returns
    -------
    estimate : float
        The Henze-Penrose estimate of the Bayes error rate

    """

    Dptilde = divergence(X1, X2)
    estimate = 1.0/2.0 - 1.0/4.0*sqrt(Dptilde) - 1.0/4.0*Dptilde
    if normalize:
        phat = X1.shape[0] / (X1.shape[0] + X2.shape[0])
        qhat = X2.shape[0] / (X1.shape[0] + X2.shape[0])
        ratio = min(phat, qhat)
        estimate = estimate / ratio
    return estimate



# -*- coding: utf-8 -*-

"""
"""

from __future__ import division

import unittest
import numpy as np

from smartsvm.cross_connections import (binary_cross_connections, 
        ovr_cross_connections)

class CrossConnectionsTestCase(unittest.TestCase):

    def test_binary_cross_connections_1(self):
        """ CROSS-CONNECTIONS: Test binary cross connections (1) """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            [4, 1],
            [3, -3],
            [5, 4]
            ])
        y = np.array([-1, -1, -1, 1, 1, 1])

        C = binary_cross_connections(X, y, nTrees=3)
        self.assertAlmostEqual(C, 2.0)

    def test_binary_cross_connections_2(self):
        """ CROSS-CONNECTIONS: Test binary cross connections (2) """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            [4, 1],
            [3, -3],
            [5, 4]
            ])
        y = np.array([-1, -1, -1, 1, 1, 1])

        C = binary_cross_connections(X, y, nTrees=4)
        self.assertAlmostEqual(C, 2.0)

    def test_ovr_cross_connections_1(self):
        """ CROSS-CONNECTIONS: Test OvR cross connections (1) """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            [4, 1],
            [3, -3],
            [5, 4]
            ])
        y = np.array([0, 0, 0, 1, 1, 2])

        C = ovr_cross_connections(X, y, 3, nTrees=1)
        self.assertTrue(np.array_equal([0, 1, 0], C))

    def test_ovr_cross_connections_2(self):
        """ CROSS-CONNECTIONS: Test OvR cross connections (2) """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            [4, 1],
            [3, -3],
            [5, 4]
            ])
        y = np.array([0, 0, 0, 1, 1, 2])

        C = ovr_cross_connections(X, y, 3, nTrees=4)
        self.assertTrue(np.array_equal([2, 5/3, 2/3], C))

    def test_ovr_cross_connections_error(self):
        """ CROSS-CONNNECTIONS: Test OvR cross connections error """
        X = np.array([
            [-1, -1],
            [-2, 4],
            [-5, -5],
            [4, 1],
            [3, -3],
            [5, 4]
            ])
        y = np.array([0, 0, 0, 1, 1, 2])
        y += 1

        with self.assertRaises(ValueError) as err:
            C = ovr_cross_connections(X, y, 3, nTrees=4)
            del C

        the_exception = err.exception
        self.assertEqual(the_exception.args,
                ("Class labels should be in {0, ..., nClass-1}",))



if __name__ == '__main__':
    unittest.main()



"""
    Unit tests for the base classes of the Hierarchical Classifiers.

    Note that since the base classes are abstract classes, we mock the abstract 
    methods and test only the ones that remain.

"""

from __future__ import division

import math
import networkx as nx
import numpy as np
import six
import unittest

from sklearn.svm import LinearSVC
from sklearn.utils.validation import NotFittedError
from smartsvm import base

if six.PY2:
    from mock import patch, PropertyMock
else:
    from unittest.mock import patch, PropertyMock


class MockClassifier(object):

    def __init__(self):
        self._X = None
        self._y = None

    def fit(self, X, y):
        self._X = X
        self._y = y
        self.coef_ = None

    def predict(self, X):
        return np.tile(1, (X.shape[0], ))

class MockHierarchicalElements(base.HierarchicalClassifier):

    def __init__(self, elements=None):
        self._elements = elements

    @property
    def elements(self):
        return self._elements


class HierarchicalClassifierTestCase(unittest.TestCase):

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set())
    def test_init_1(self):
        """ BASE: HierarchicalClassifier test init defaults """
        inst = base.HierarchicalClassifier()
        self.assertEqual(inst.binary_clf, LinearSVC)
        self.assertEqual(inst.n_jobs, 1)
        self.assertEqual(inst.clf_params, {})

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set())
    def test_init_2(self):
        """ BASE: HierarchicalClassifier test init specified """
        clf = MockClassifier()
        inst = base.HierarchicalClassifier(binary_clf=clf, n_jobs=2,
                clf_params={'C': 10.0})
        self.assertEqual(inst.binary_clf, clf)
        self.assertEqual(inst.n_jobs, 2)
        self.assertEqual(inst.clf_params, {'C': 10.0})

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            elements=PropertyMock(return_value=[10]))
    def test_is_splitable_1(self):
        """ BASE: HierarchicalClassifier test is_splitable (1) """
        inst = base.HierarchicalClassifier()
        self.assertFalse(inst.is_splitable)

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            elements=PropertyMock(return_value=[10, 10]))
    def test_is_splitable_2(self):
        """ BASE: HierarchicalClassifier test is_splitable (2) """
        inst = base.HierarchicalClassifier()
        self.assertTrue(inst.is_splitable)

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set())
    def test_is_splitted(self):
        """ BASE: HierarchicalClassifier test is_splitted """
        inst = base.HierarchicalClassifier()
        self.assertFalse(inst.is_splitted)
        inst.positive_ = None
        self.assertFalse(inst.is_splitted)
        inst.negative_ = None
        self.assertTrue(inst.is_splitted)

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            is_splitable=PropertyMock(return_value=False),
            elements=PropertyMock(return_value=[10]))
    def test_predict_1(self):
        """ BASE: HierarchicalClassifier test predict not splitable """
        inst = base.HierarchicalClassifier()
        X = np.array([[1, 1], [2, 2]])
        res = inst.predict(X)
        exp = np.array([10, 10])
        self.assertEqual(exp[0], res[0])
        self.assertEqual(exp[1], res[1])

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            is_splitable=PropertyMock(return_value=True))
    def test_predict_2(self):
        """ BASE: HierarchicalClassifier test predict empty data """
        inst = base.HierarchicalClassifier()
        X = np.array(())
        res = inst.predict(X)
        exp = np.array(())
        self.assertTrue(np.array_equal(exp, res))

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            is_splitable=PropertyMock(return_value=True))
    def test_predict_3(self):
        """ BASE: HierarchicalClassifier test predict not fitted (1) """
        inst = base.HierarchicalClassifier()
        X = np.array([[1, 1], [2, 2]])
        with self.assertRaises(NotFittedError) as cm:
            res = inst.predict(X)
            del res

        the_exception = cm.exception
        self.assertEqual(the_exception.args,
                ("This HierarchicalClassifier instance is not fitted yet. "
                    "Call 'fit' with appropriate arguments before using "
                    "this method.", ))

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            elements=PropertyMock(return_value=[10, 5]))
    @patch.multiple(MockHierarchicalElements, __abstractmethods__=set())
    def test_predict_5(self):
        """ BASE: HierarchicalClassifier test predict """
        inst = base.HierarchicalClassifier()
        X = np.array([[1, 1], [2, 2]])
        inst.classifier_ = MockClassifier()
        inst.classifier_.fit(X, None)
        inst.negative_child_ = MockHierarchicalElements(elements=[10])
        inst.positive_child_ = MockHierarchicalElements(elements=[5])
        res = inst.predict(X)
        exp = np.array([5, 5])
        self.assertTrue(np.array_equal(res, exp))

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            is_splitable=PropertyMock(return_value=False))
    def test_set_binary_1(self):
        """ BASE: HierarchicalClassifier test _set_binary not splitable """
        inst = base.HierarchicalClassifier()
        inst._set_binary()
        self.assertEqual(inst.classifier_, None)

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            is_splitable=PropertyMock(return_value=True))
    def test_set_binary_2(self):
        """ BASE: HierarchicalClassifier test _set_binary with clf"""
        inst = base.HierarchicalClassifier()
        clf = MockClassifier()
        inst._set_binary(clf=clf)
        self.assertEqual(inst.classifier_, clf)

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            is_splitable=PropertyMock(return_value=True))
    def test_set_binary_3(self):
        """ BASE: HierarchicalClassifier test _set_binary without clf (1)"""
        inst = base.HierarchicalClassifier()
        inst._set_binary()
        self.assertTrue(isinstance(inst.classifier_, LinearSVC))

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            is_splitable=PropertyMock(return_value=True))
    def test_set_binary_4(self):
        """ BASE: HierarchicalClassifier test _set_binary without clf (2)"""
        inst = base.HierarchicalClassifier(binary_clf=MockClassifier)
        inst._set_binary()
        self.assertTrue(isinstance(inst.classifier_, MockClassifier))

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            is_splitable=PropertyMock(return_value=True))
    def test_get_binary_1(self):
        """ BASE: HierarchicalClassifier test _get_binary (1) """
        inst = base.HierarchicalClassifier()
        clf = inst._get_binary()
        self.assertTrue(isinstance(clf, LinearSVC))

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            is_splitable=PropertyMock(return_value=True))
    def test_get_binary_2(self):
        """ BASE: HierarchicalClassifier test _get_binary (2) """
        inst = base.HierarchicalClassifier()
        inst.classifier_ = MockClassifier()
        clf = inst._get_binary()
        self.assertTrue(isinstance(clf, MockClassifier))

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set())
    def test_binary_data_1(self):
        """ BASE: HierarchicalClassifier test _binary_data (1)"""
        inst = base.HierarchicalClassifier()
        inst.negative_ = [4]
        inst.positive_ = [2, 3]
        X = np.array([
            [1.0, 1.0, 1.0],
            [1.5, 1.5, 1.5],
            [2.0, 2.0, 2.0],
            [2.5, 2.5, 2.5],
            [3.0, 3.0, 3.0],
            [3.5, 3.5, 3.5],
            [4.0, 4.0, 4.0],
            [4.5, 4.5, 4.5]
            ])
        y = np.array([1, 1, 2, 2, 3, 3, 4, 4])

        res_X, res_y = inst._binary_data(X, y)
        exp_X = np.array([
            [2.0, 2.0, 2.0],
            [2.5, 2.5, 2.5],
            [3.0, 3.0, 3.0],
            [3.5, 3.5, 3.5],
            [4.0, 4.0, 4.0],
            [4.5, 4.5, 4.5]
            ])
        exp_y = np.array([1, 1, 1, 1, -1, -1])
        self.assertTrue(np.array_equal(res_X, exp_X))
        self.assertTrue(np.array_equal(res_y, exp_y))

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set())
    def test_binary_data_2(self):
        """ BASE: HierarchicalClassifier test _binary_data (2)"""
        inst = base.HierarchicalClassifier()
        inst.negative_ = [8]
        inst.positive_ = [2, 3]
        X = np.array([
            [1.0, 1.0, 1.0],
            [1.5, 1.5, 1.5],
            [2.0, 2.0, 2.0],
            [2.5, 2.5, 2.5],
            [3.0, 3.0, 3.0],
            [3.5, 3.5, 3.5],
            [4.0, 4.0, 4.0],
            [4.5, 4.5, 4.5]
            ])
        y = np.array([1, 1, 2, 2, 3, 3, 4, 4])

        res_X, res_y = inst._binary_data(X, y)
        exp_X = np.array([
            [2.0, 2.0, 2.0],
            [2.5, 2.5, 2.5],
            [3.0, 3.0, 3.0],
            [3.5, 3.5, 3.5],
            ])
        exp_y = np.array([1, 1, 1, 1])
        self.assertTrue(np.array_equal(res_X, exp_X))
        self.assertTrue(np.array_equal(res_y, exp_y))

    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set())
    def test_binary_data_3(self):
        """ BASE: HierarchicalClassifier test _binary_data (3)"""
        inst = base.HierarchicalClassifier()
        inst.negative_ = ['orange']
        inst.positive_ = ['blue', 'green']
        X = np.array([
            [1.0, 1.0, 1.0],
            [1.5, 1.5, 1.5],
            [2.0, 2.0, 2.0],
            [2.5, 2.5, 2.5],
            [3.0, 3.0, 3.0],
            [3.5, 3.5, 3.5],
            [4.0, 4.0, 4.0],
            [4.5, 4.5, 4.5]
            ])
        y = np.array(['red', 'red', 'blue', 'blue', 'green', 'green', 'purple', 
            'purple'])

        res_X, res_y = inst._binary_data(X, y)
        exp_X = np.array([
            [2.0, 2.0, 2.0],
            [2.5, 2.5, 2.5],
            [3.0, 3.0, 3.0],
            [3.5, 3.5, 3.5],
            ])
        exp_y = np.array([1, 1, 1, 1])
        self.assertTrue(np.array_equal(res_X, exp_X))
        self.assertTrue(np.array_equal(res_y, exp_y))


    @patch.multiple(base.HierarchicalClassifier, __abstractmethods__=set(),
            is_splitable=PropertyMock(return_value=True))
    def test_fit_binary_1(self):
        """ BASE: HierarchicalClassifier test fit_binary """
        inst = base.HierarchicalClassifier(binary_clf=MockClassifier)
        inst.negative_ = [4]
        inst.positive_ = [2, 3]
        X = np.array([
            [1.0, 1.0, 1.0],
            [1.5, 1.5, 1.5],
            [2.0, 2.0, 2.0],
            [2.5, 2.5, 2.5],
            [3.0, 3.0, 3.0],
            [3.5, 3.5, 3.5],
            [4.0, 4.0, 4.0],
            [4.5, 4.5, 4.5]
            ])
        y = np.array([1, 1, 2, 2, 3, 3, 4, 4])

        exp_X = np.array([
            [2.0, 2.0, 2.0],
            [2.5, 2.5, 2.5],
            [3.0, 3.0, 3.0],
            [3.5, 3.5, 3.5],
            [4.0, 4.0, 4.0],
            [4.5, 4.5, 4.5]
            ])
        exp_y = np.array([1, 1, 1, 1, -1, -1])

        inst._fit_binary(X, y)
        self.assertTrue(hasattr(inst, "classifier_"))
        self.assertTrue(hasattr(inst.classifier_, "coef_"))
        self.assertTrue(np.array_equal(inst.classifier_._X, exp_X))
        self.assertTrue(np.array_equal(inst.classifier_._y, exp_y))


def mock_split(obj):
    print("spliting on obj: %r" % id(obj))
    exp_weight = 1.0/2.0 - 1.0/4.0 * math.sqrt(1.0/3.0) - 1.0/4.0*1.0/3.0
    if len(obj.elements) == 3:
        obj.negative_ = set([1, 2])
        obj.positive_ = set([3])
        G_neg = nx.Graph()
        G_neg.add_edge(1, 2, weight=exp_weight)
        G_pos = nx.Graph()
        G_pos.add_node(3)
    elif len(obj.elements) == 2:
        obj.negative_ = set([1])
        obj.positive_ = set([2])
        G_neg = nx.Graph()
        G_neg.add_node(1)
        G_pos = nx.Graph()
        G_pos.add_node(2)
    obj.negative_child_ = base.GraphClassifier(binary_clf=MockClassifier, 
            graph=G_neg)
    obj.positive_child_ = base.GraphClassifier(binary_clf=MockClassifier, 
            graph=G_pos)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Base object for hierarchical classifiers

This module contains the definition of a general hierarchical classifier, which 
forms the basis for the :class:`.SmartSVM` classifier. The 
:class:`.HierarchicalClassifier` class is an abstract base class, which leaves 
the ``fit`` and ``_split`` methods to be defined in the subclass.

"""

from __future__ import division

import numpy as np
import six

from abc import ABCMeta, abstractmethod, abstractproperty

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.svm import LinearSVC
from sklearn.utils.validation import check_is_fitted

from .utils import indices_from_classes


@six.add_metaclass(ABCMeta)
class HierarchicalClassifier(BaseEstimator, ClassifierMixin):
    """Base class for a hierarchical classifier

    This class is a base class for a hierarchical classifier which contains a 
    hierarchy of binary classification problems. It forms the basis of the 
    :class:`.SmartSVM` class and can be used to implement other hierarchical 
    classifiers. Note that this class is also the node class for the binary 
    tree: it has children which are in turn also 
    :class:`.HierarchicalClassifier` instances.

    Parameters
    ----------

    binary_clf : classifier
        The type of the binary classifier to be used for each binary 
        subproblem. This will typically be a scikit-learn classifier such as 
        LinearSVC, DecisionTreeClassifier, etc.

    clf_params : dict
        Parameters to pass on to the constructor of the ``binary_clf`` type 
        classifier. It must be a dict with a mapping from parameter to value.

    n_jobs : int
        Number of parallel jobs to use where applicable.


    Attributes
    ----------

    classifier_ : classifier
        The binary classifier for this node in the tree, if it exists. This can 
        also be obtained with the ``_get_binary()`` method.

    negative_child_ : HierarchicalClassifier
        The "left" child of the binary tree below this node.

    positive_child_ : HierarchicalClassifier
        The "right" child of the binary tree below this node.

    negative_ : set
        Set of labels of the binary subproblem that are on the "left" part of 
        the tree.

    positive_ : set
        Set of labels on the binary subproblem that are on the "right" part of 
        the tree.

    Methods
    -------

    predict(X)
        Perform classification on dataset X

    """

    def __init__(self, binary_clf=LinearSVC, clf_params=None, n_jobs=1):
        self.binary_clf = binary_clf
        self.clf_params = {} if clf_params is None else clf_params
        self.n_jobs = n_jobs

    @abstractmethod # pragma: no cover
    def fit(self, X, y):
        """ Fit model. """

    @abstractproperty # pragma: no cover
    def elements(self):
        """ Elements in the graph or list of nodes """

    @abstractmethod # pragma: no cover
    def _split(self):
        """ Method for splitting up the classes. """

    def predict(self, X):
        """ Predict the class labels using the hierarchical classifier """
        if X.shape[0] == 0:
            return np.tile(float("nan"), (X.shape[0], ))
        if not self.is_splitable:
            return np.tile(self.elements[0], (X.shape[0], ))
        check_is_fitted(self, "classifier_")
        check_is_fitted(self, "negative_child_")
        check_is_fitted(self, "positive_child_")
        pred = self.classifier_.predict(X)
        idx_neg = pred == -1
        idx_pos = pred == 1
        dtype = np.array(self.elements).dtype
        predictions = np.zeros((X.shape[0], ), dtype=dtype)
        predictions[idx_neg, ] = self.negative_child_.predict(X[idx_neg, :])
        predictions[idx_pos, ] = self.positive_child_.predict(X[idx_pos, :])
        return predictions

    @property
    def is_splitable(self):
        """ Check if the elements of this node can be split further """
        return len(self.elements) > 1

    @property
    def is_splitted(self):
        """ Check if this node is already split """
        return hasattr(self, "positive_") and hasattr(self, "negative_")

    def _stack(self):
        """ Construct a flat list of all the nodes in the hierarchy """
        stack = [self]
        if self.is_splitted:
            stack.extend(self.negative_child_._stack())
            stack.extend(self.positive_child_._stack())
        return stack

    def _split_all(self):
        """ Recursively split all nodes """
        self._split()
        if self.is_splitted:
            self.negative_child_._split_all()
            self.positive_child_._split_all()

    def _set_binary(self, clf=None):
        """ Initialize the binary classifier of this node """
        if not self.is_splitable:
            self.classifier_ = None
            return
        if not clf is None:
            self.classifier_ = clf
            return
        self.classifier_ = self.binary_clf(**self.clf_params)

    def _get_binary(self):
        """ Return the classifier of the node, initializing if needed """
        if not hasattr(self, "classifier_"):
            self._set_binary()
        return self.classifier_

    def _binary_data(self, X, y):
        """ Return data for the binary classification problem of the node """
        idx_negative = indices_from_classes(self.negative_, y)
        idx_positive = indices_from_classes(self.positive_, y)
        idx_all = np.logical_or(idx_negative, idx_positive)

        y_np = np.zeros((y.shape[0], ))
        y_np[idx_negative, ] = -1
        y_np[idx_positive, ] = 1

        X_needed = X[idx_all, ]
        y_needed = y_np[idx_all, ]

        return X_needed, y_needed

    def _fit_binary(self, X, y):
        """ Fit the classifier of the node """
        X_needed, y_needed = self._binary_data(X, y)
        self._set_binary()
        self.classifier_.fit(X_needed, y_needed)

    def __str__(self): # pragma: no cover
        s = repr(self) + "\n"
        s += "Tree:" + "\n"
        s += self.print_tree()
        return s

    def print_tree(self, prefix="", is_tail=None): # pragma: no cover
        """ Print the tree """
        s = prefix
        if is_tail is None:
            s += ""
        elif is_tail:
            s += "└── R "
        else:
            s += "├── L "

        s += repr(self.negative_.union(self.positive_)) + "\n"
        if is_tail is None or is_tail:
            buf = "%s    " % prefix
        else:
            buf = "%s|   " % prefix

        if self.negative_child_ and len(self.negative_) > 1:
            s += self.negative_child_.print_tree(prefix=buf, is_tail=False)
        else:
            s += buf + "├── L " + repr(self.negative_) + "\n"

        if self.positive_child_ and len(self.positive_) > 1:
            s += self.positive_child_.print_tree(prefix=buf, is_tail=True)
        else:
            s += buf + "└── R " + repr(self.positive_) + "\n"

        return s



#!/usr/bin/env python

import os
import re

#from distutils.core import setup
from setuptools import setup, find_packages
from distutils.extension import Extension

# Set this to True to enable building extensions using Cython. Set it to False 
# to build extensions from the C file (that was previously generated using 
# Cython). Set it to 'auto' to build with Cython if available, otherwise from 
# the C file.
USE_CYTHON = 'auto'

if USE_CYTHON:
    try:
        from Cython.Distutils import build_ext
    except ImportError:
        if USE_CYTHON == 'auto':
            USE_CYTHON = False
        else:
            raise

cmdclass = {}
ext_modules = []

if USE_CYTHON:
    ext_modules += [
            Extension("smartsvm.ortho_mst", [
                    os.path.join("src", "ortho_mst.pyx"),
                    os.path.join("src", "c_ortho_mst.c")
                    ]),
            Extension("smartsvm.multiclass_mst_count", [
                os.path.join("src", "multiclass_mst_count.pyx"),
                os.path.join("src", "c_multiclass_mst_count.c")
                ])
            ]
    cmdclass.update({'build_ext': build_ext})
else:
    ext_modules += [
            Extension("smartsvm.ortho_mst", [
                os.path.join("src", "ortho_mst.c"),
                os.path.join("src", "c_ortho_mst.c")
                ]),
            Extension("smartsvm.multiclass_mst_count", [
                os.path.join("src", "multiclass_mst_count.c"),
                os.path.join("src", "c_multiclass_mst_count.c")
                ])
            ]

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


version = re.search("__version__ = '([^']+)'", 
        open('smartsvm/__init__.py').read()).group(1)

setup(
        name='smartsvm',
        version=version,
        description=("Python package for Meta-Learning and Adaptive "
            "Hierarchical Classifier Design"),
        long_description=read('README.rst'),
        url="https://github.com/HeroResearchGroup/SmartSVM",
        author='G.J.J. van den Burg',
        author_email='gertjanvandenburg@gmail.com',
        license='GPL v2',
        install_requires=[
            'networkx>=1.9',
            'scipy',
            'scikit-learn',
            'numpy'
            ],
        packages=find_packages(),
        cmdclass=cmdclass,
        ext_modules=ext_modules,
        classifiers=[
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            ]
)



"""
    Unit tests for the cut module of the SmartSVM package.
"""

from __future__ import division

import unittest
import networkx as nx

from smartsvm import cut


class CutTestCase(unittest.TestCase):

    def test_graph_cut_1(self):
        """ CUT: Test main graph_cut function - default args"""
        # This function simply tests if dispatching is done correctly.
        G = nx.Graph()
        G.add_edge('x','y', weight=3)
        G1_exp = nx.Graph()
        G1_exp.add_node('x')
        G2_exp = nx.Graph()
        G2_exp.add_node('y')

        G1_res, G2_res = cut.graph_cut(G)
        self.assertTrue((sorted(G1_res.nodes()) == sorted(G1_exp.nodes())) ^
                (sorted(G1_res.nodes()) == sorted(G2_exp.nodes())))
        self.assertTrue((sorted(G2_res.nodes()) == sorted(G1_exp.nodes())) ^
                (sorted(G2_res.nodes()) == sorted(G2_exp.nodes())))

    def test_graph_cut_2(self):
        """ CUT: Test main graph_cut function - Stoer-Wagner"""
        G = nx.Graph()
        G.add_edge('x', 'y', weight=3)
        G1_exp = nx.Graph()
        G1_exp.add_node('x')
        G2_exp = nx.Graph()
        G2_exp.add_node('y')

        G1_res, G2_res = cut.graph_cut(G, algorithm='stoer_wagner')
        self.assertTrue((sorted(G1_res.nodes()) == sorted(G1_exp.nodes())) ^
                (sorted(G1_res.nodes()) == sorted(G2_exp.nodes())))
        self.assertTrue((sorted(G2_res.nodes()) == sorted(G1_exp.nodes())) ^
                (sorted(G2_res.nodes()) == sorted(G2_exp.nodes())))

    def test_graph_cut_3(self):
        """ CUT: Test main graph_cut function - Spectral Clustering (1) """
        G = nx.Graph()
        G.add_edge('x', 'y', weight=3)
        G.add_edge('x', 'z', weight=5)
        G1_exp = nx.Graph()
        G1_exp.add_node('x')
        G1_exp.add_node('y')
        G2_exp = nx.Graph()
        G2_exp.add_node('z')

        G1_res, G2_res = cut.graph_cut(G, algorithm='spectral_clustering')
        self.assertTrue((sorted(G1_res.nodes()) == sorted(G1_exp.nodes())) ^
                (sorted(G1_res.nodes()) == sorted(G2_exp.nodes())))
        self.assertTrue((sorted(G2_res.nodes()) == sorted(G1_exp.nodes())) ^
                (sorted(G2_res.nodes()) == sorted(G2_exp.nodes())))

    def test_graph_cut_4(self):
        """ CUT: Test main graph_cut function - Spectral Clustering (2) """
        G = nx.Graph()
        G.add_edge('x', 'y', weight=3)
        G.add_edge('x', 'z', weight=5)
        G1_exp = nx.Graph()
        G1_exp.add_node('x')
        G1_exp.add_node('y')
        G2_exp = nx.Graph()
        G2_exp.add_node('z')

        G1_res, G2_res = cut.graph_cut(G, algorithm='normalized_cut')
        self.assertTrue((sorted(G1_res.nodes()) == sorted(G1_exp.nodes())) ^
                (sorted(G1_res.nodes()) == sorted(G2_exp.nodes())))
        self.assertTrue((sorted(G2_res.nodes()) == sorted(G1_exp.nodes())) ^
                (sorted(G2_res.nodes()) == sorted(G2_exp.nodes())))

    def test_graph_cut_5(self):
        """ CUT: Test main graph_cut function - Unknown algorithm"""
        # This function simply tests if dispatching is done correctly.
        G = nx.Graph()
        G.add_edge('x','y', weight=3)
        with self.assertRaises(ValueError) as cm:
            cut.graph_cut(G, algorithm='test')

        the_exception = cm.exception
        self.assertEqual(the_exception.args,
                ("Unknown graph cut algorithm supplied: test",))

    def test_graph_cut_sw(self):
        """ CUT: Test Stoer-Wagner graph cut. """
        # This test verifies if the returned graphs are the expected subgraphs.
        # Uses the example of Stoer-Wagner from the networkx docs, slightly 
        # modified to make sure the cut is unique.
        G = nx.Graph()
        G.add_edge('x','a', weight=3)
        G.add_edge('x','b', weight=2)
        G.add_edge('a','c', weight=3)
        G.add_edge('b','c', weight=5)
        G.add_edge('b','d', weight=4)
        G.add_edge('d','e', weight=2)
        G.add_edge('c','y', weight=2)
        G.add_edge('e','y', weight=3)
        G1_exp = nx.Graph()
        G1_exp.add_edge('x','a', weight=3)
        G1_exp.add_edge('x','b', weight=2)
        G1_exp.add_edge('a','c', weight=3)
        G1_exp.add_edge('b','c', weight=5)
        G1_exp.add_edge('b','d', weight=4)
        G2_exp = nx.Graph()
        G2_exp.add_edge('e','y', weight=3)

        G1_res, G2_res = cut.graph_cut_sw(G)
        em = nx.algorithms.isomorphism.numerical_edge_match('weight', 1)
        self.assertTrue(nx.is_isomorphic(G1_res, G1_exp, edge_match=em) ^
                nx.is_isomorphic(G1_res, G2_exp, edge_match=em))
        self.assertTrue(nx.is_isomorphic(G2_res, G1_exp, edge_match=em) ^
                nx.is_isomorphic(G2_res, G2_exp, edge_match=em))
        self.assertTrue((sorted(G1_res.nodes()) == sorted(G1_exp.nodes())) ^
                (sorted(G1_res.nodes()) == sorted(G2_exp.nodes())))
        self.assertTrue((sorted(G2_res.nodes()) == sorted(G1_exp.nodes())) ^
                (sorted(G2_res.nodes()) == sorted(G2_exp.nodes())))


if __name__ == '__main__':
    unittest.main()
